// =========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_format.h,v $
// (c) INRIA, projet APACHE, 2001
// Author: T. Gautier
//
//
// =========================================================================
#ifndef _ACOM_FORMAT_H_
#define _ACOM_FORMAT_H_

#include <inuk/acom_netconfig.h>
#include <inuk/acom_error.h>
#include <inuk/acom_stream.h>
#include <vector>

/* Un fichier les envois/receptions de donnees
*/
namespace ACOM_NET_NAMESPACE { 

#ifndef INUK_DEBUG_MEM
#define acom_inuk_malloc( s) ::malloc(s)
#define acom_inuk_calloc( s1, s2) ::calloc(s1,s2)
#define acom_inuk_free( p ) ::free(p)
#else
#define acom_inuk_malloc( s) _acom_inuk_malloc(s)
#define acom_inuk_calloc( s1, s2) _acom_inuk_calloc(s1,s2)
#define acom_inuk_free( p ) _acom_inuk_free(p)
extern void* _acom_inuk_malloc( size_t s);
extern void* _acom_inuk_calloc( size_t s1, size_t s2);
extern void  _acom_inuk_free( void* p );
#endif

class OStream;
class IStream;
class StreamFormat;

// --------------------------------------------------------------------
/*
 * A Format is the definition of the format for a data structure,
 * currently only XDR like format is supported. MPI like format will
 * be supported as a compilation of XDR like format.
 */
class Format {
public:
  // - signature for arguments of the following methods
  typedef void (Format::*f_write)( OStream& o, const void* d, size_t count ) const;
  typedef void (Format::*f_read) ( IStream& o, void* d, size_t count ) const;
  typedef void* (Format::*f_allocate) ( size_t count ) const;
  typedef void (Format::*f_deallocate) (void* d, size_t count) const;
  typedef void (Format::*f_cstor) (void* d, size_t count) const;
  typedef void (Format::*f_dstor) (void* d, size_t count) const;
  typedef void (Format::*f_copy) ( void* dest, const void* src, size_t count ) const;
  typedef void* (Format::*f_create) ( size_t count ) const;
  typedef void (Format::*f_destroy) (void* d, size_t count ) const;

  // - type of a format identifier
  typedef long Id;

  // - default cstor
  Format( );

  // - cstor with parameter to define a format
  Format( size_t sz, f_write fwrite, f_read fread,
          f_allocate falloc, f_deallocate fdealloc,
          f_cstor fcstor, f_dstor fdstor,
          f_create fcreate, f_destroy fdestroy,
          f_copy fcopy );

  // - cstor with parameter to define a format
  Format( const std::string& n, 
          size_t sz, f_write fwrite, f_read fread, 
          f_allocate falloc, f_deallocate fdealloc, 
          f_cstor fcstor, f_dstor fdstor, 
          f_create fcreate, f_destroy fdestroy, 
          f_copy fcopy);

  // - recopy cstor: make an alias
  Format( const Format& fmt );

  // - assignment operator : make an alias
  Format& operator=( const Format& fmt );


  // - return the format with identifier 'id'
  static Format* get_format( Id id )
  { return _all_fm[id]; }

  // - return the identifier of the format 
  Id get_id( ) const
  { return _id; }

  // - return the size in bytes of a C++ data with format 
  size_t get_size( ) const
  { return _size; }

  // Apply the different methods 
  void write( OStream& o, const void* d, size_t count =1 ) const
  { (this->*_fwrite)(o, d, count); }

  void read( IStream& o, void* d, size_t count =1 ) const
  { (this->*_fread)(o, d, count); }

  void* allocate(size_t count =1) const 
  { return (this->*_falloc)(count); }

  void  deallocate(void* d, size_t count =1) const 
  { (this->*_fdealloc)(d, count); }

  void  cstor(void* d, size_t count =1) const 
  { (this->*_fcstor)(d, count); }

  void  dstor(void* d, size_t count =1) const 
  { (this->*_fdstor)(d, count); }

  void* create(size_t count =1) const 
  { return (this->*_fcreate)(count); }

  void  destroy(void* d, size_t count =1) const 
  { (this->*_fdestroy)(d, count); }

  void  copy(void* dest, const void* src, size_t count =1) const 
  { (this->*_fcopy)(dest, src, count); }


  // Return the functions 
  f_write      get_write() const { return _fwrite; }
  f_read       get_read() const { return _fread; }
  f_allocate   get_allocate() const { return _falloc; }
  f_deallocate get_deallocate() const { return _fdealloc; }
  f_cstor      get_cstor() const { return _fcstor; }
  f_dstor      get_dstor() const { return _fdstor; }
  f_create     get_create() const { return _fcreate; }
  f_destroy    get_destroy() const { return _fdestroy; }
  f_copy       get_copy() const { return _fcopy; }

  // - Set the size in bytes of a C++ data with format 
  void set_size( size_t s ) 
  { _size =s; }

  const std::string& get_name() const
  { return _name; }

  void get_name(const std::string& s)
  { _name = s; }

  /** Return true iff the format has immediate data
  */
  bool has_immediate() const
  { return _hasimmediate; }

  /** Return true iff the format is contiguous
  */
  bool is_contiguous() const
  { return _iscontiguous; }

  /** Compile a format 
  */
  void compile() const;
  
  /** Compile all registered not yet compiled formats 
  */
  static void compile_all();

protected:
  static void init_format();
  friend class ObjectInitFormat;
  friend class StreamFormat;

protected:
  Id           _id;
  std::string  _name;
  size_t       _size;
  f_write      _fwrite;
  f_read       _fread;
  f_allocate   _falloc;
  f_deallocate _fdealloc;
  f_cstor      _fcstor;
  f_dstor      _fdstor;
  f_create     _fcreate;
  f_destroy    _fdestroy;
  f_copy       _fcopy;

  mutable bool _iscompiled;
  mutable bool _couldbecompiled;
  mutable bool _iscontiguous;
  mutable bool _hasimmediate;

  struct OffSetDataOut {
    OffSetDataOut( const Format* f =0, long o=0, long c=0 )
     : fmt(f), offset(o), count(c)
    {}
    const Format* fmt;
    long          offset;
    long          count;
  };
  mutable std::vector<OffSetDataOut> _out_iov_i;
  mutable std::vector<OffSetDataOut> _out_iov_d;
  
private:
  static int     _counter;   // - counter to numbering format
  static Format* _all_fm[];  // - array of format
};


// --------------------------------------------------------------------
namespace FormatDef {
  extern Format Null;
  extern const Format Bool;
  extern const Format Byte;
  extern const Format Char;
  extern const Format UChar;
  extern const Format Int;
  extern const Format UInt;
  extern const Format Short;
  extern const Format UShort;
  extern const Format Long;
  extern const Format ULong;
  extern const Format LLong;
  extern const Format ULLong;
  extern const Format Float;
  extern const Format Double;
  extern const Format LDouble;
};

  
// --------------------------------------------------------------------
class ObjectInitFormat {
public:
  ObjectInitFormat();
};

static ObjectInitFormat TheInitializor;


// --------------------------------------------------------------------
template<class T>
class ACOMWrapperFormat : public ACOM::Format {
public:
  void write( OStream& s, const void* val, size_t count ) const;

  void read( IStream& s, void* val, size_t count ) const;

  void* allocate( size_t count ) const
  { return acom_inuk_malloc(count*sizeof(T)); }
  
  void deallocate( void* d, size_t count  ) const
  { acom_inuk_free(d); }
  
  void cstor( void* d, size_t count ) const
  { T* dd = (T*)d; 
    for (size_t i=0; i<count; ++i) 
      new (&dd[i]) T; 
  }

  void dstor( void* d, size_t count ) const
  { T* dd = (T*)d; 
    for (size_t i=0; i<count; ++i) 
      dd[i].~T(); 
  }

  void* create( size_t count ) const
  { T*d = (T*)acom_inuk_malloc(count*sizeof(T)); 
    for (size_t i=0; i<count; ++i) 
      new (&d[i]) T; 
    return d; 
  }

  void destroy( void* d, size_t count ) const
  { T* dd = (T*)d; for (size_t i=0; i<count; ++i) dd[i].~T(); acom_inuk_free(d); }

  void copy( void* dest, const void* src, size_t count ) const
  { T* ddest = (T*)dest; 
    const T* dsrc = (const T*)src;
    for (size_t i=0; i<count; ++i)
      ddest[i] = dsrc[i];
  }

  static const Format theformat;
  static const Format* const format;
  static const Format* get_format();
};



// --------------------------------------------------------------------
#ifndef PTH_USE_PTHREAD_IRIX
#define ACOM_SPECIALIZED_FORMAT(TYPE,OBJ)\
template<>\
class ACOMWrapperFormat<TYPE> : public ACOM::Format {\
public:\
  void write( OStream& s, const void* val, size_t count ) const\
  { s.write( format, OStream::DA, val, count ); }\
  void read( IStream& s, void* val, size_t count ) const\
  { s.read( format, OStream::DA, val, count ); }\
  void* allocate( size_t count ) const\
  { return acom_inuk_malloc(sizeof(TYPE)*count); }\
  void deallocate( void* d, size_t count ) const\
  { char* dd = (char*)d; acom_inuk_free(dd); }\
  void cstor( void* d, size_t count ) const\
  { new (d) TYPE[count]; }\
  void dstor( void* d, size_t count ) const\
  { }\
  void* create( size_t count ) const\
  { void*d = acom_inuk_malloc(sizeof(TYPE)*count); new (d) TYPE[count]; return d; }\
  void destroy( void* d, size_t count ) const\
  { acom_inuk_free(d); }\
  void copy( void* dest, const void* src, size_t count ) const\
  { TYPE* ddest = (TYPE*)dest; \
    const TYPE* dsrc = (const TYPE*)src;\
    for (size_t i=0; i<count; ++i) \
      ddest[i] = dsrc[i];\
  }\
  static const Format& theformat;\
  static const Format* const format;\
  static const Format* get_format()\
  { return format; }\
};
#else
#define ACOM_SPECIALIZED_FORMAT(TYPE,OBJ)\
template<>\
class ACOMWrapperFormat<TYPE> : public ACOM::Format {\
public:\
  void write( OStream& s, const void* val, size_t count ) const\
  { s.write( format, OStream::DA, val, count ); }\
  void read( IStream& s, void* val, size_t count ) const\
  { s.read( format, OStream::DA, val, count ); }\
  void* allocate( size_t count ) const\
  { return acom_inuk_malloc(sizeof(TYPE)*count); }\
  void deallocate( void* d, size_t count ) const\
  { acom_inuk_free(d); }\
  void cstor( void* d, size_t count ) const\
  { TYPE*dd=(TYPE*)d; for (size_t i=0; i<count; ++i) new (&dd[i]) TYPE; }\
  void dstor( void* d, size_t count ) const\
  { }\
  void* create( size_t count ) const\
  { TYPE*d=(TYPE*)acom_inuk_malloc(sizeof(TYPE)*count); for (size_t i=0; i<count; ++i) new (&d[i]) TYPE; return d; }\
  void destroy( void* d, size_t count ) const\
  { acom_inuk_free(d); }\
  void copy( void* dest, const void* src, size_t count ) const\
  { TYPE* ddest = (TYPE*)dest; \
    const TYPE* dsrc = (const TYPE*)src;\
    for (size_t i=0; i<count; ++i) \
      ddest[i] = dsrc[i];\
  }\
  static const Format& theformat;\
  static const Format* const format;\
  static const Format* get_format()\
  { return format; }\
};
#endif


ACOM_SPECIALIZED_FORMAT(bool, FormatDef::Bool)
ACOM_SPECIALIZED_FORMAT(char, FormatDef::Char)
ACOM_SPECIALIZED_FORMAT(unsigned char, FormatDef::UChar)
ACOM_SPECIALIZED_FORMAT(int, FormatDef::Int)
ACOM_SPECIALIZED_FORMAT(unsigned int, FormatDef::UInt)
ACOM_SPECIALIZED_FORMAT(short, FormatDef::Short)
ACOM_SPECIALIZED_FORMAT(unsigned short, FormatDef::UShort)
ACOM_SPECIALIZED_FORMAT(long, FormatDef::Long)
ACOM_SPECIALIZED_FORMAT(unsigned long, FormatDef::ULong)
ACOM_SPECIALIZED_FORMAT(long long, FormatDef::LLong)
ACOM_SPECIALIZED_FORMAT(unsigned long long, FormatDef::ULLong)
ACOM_SPECIALIZED_FORMAT(float, FormatDef::Float)
ACOM_SPECIALIZED_FORMAT(double, FormatDef::Double)
ACOM_SPECIALIZED_FORMAT(long double, FormatDef::LDouble)


// --------------------------------------------------------------------
template<class T>
const Format ACOMWrapperFormat<T>::theformat(
       sizeof(T),
       (Format::f_write)&ACOMWrapperFormat<T>::write,
       (Format::f_read)&ACOMWrapperFormat<T>::read,
       (Format::f_allocate)&ACOMWrapperFormat<T>::allocate,
       (Format::f_deallocate)&ACOMWrapperFormat<T>::deallocate,
       (Format::f_cstor)&ACOMWrapperFormat<T>::cstor,
       (Format::f_dstor)&ACOMWrapperFormat<T>::dstor,
       (Format::f_create)&ACOMWrapperFormat<T>::create,
       (Format::f_destroy)&ACOMWrapperFormat<T>::destroy,
       (Format::f_copy)&ACOMWrapperFormat<T>::copy
     );

template<class T>
const Format* const ACOMWrapperFormat<T>::format 
  = &ACOMWrapperFormat<T>::theformat;

template<class T>
inline const Format* ACOMWrapperFormat<T>::get_format()
{ 
  return ACOMWrapperFormat<T>::format; 
}

} // namespace ACOM_NET_NAMESPACE 


// --------------------------------------------------------------------
template<class T>
void ACOM_NET_NAMESPACE::ACOMWrapperFormat<T>::write
 ( ACOM_NET_NAMESPACE::OStream& s, const void* val, size_t count ) const
{
  const T* ref = (const T*)val;
  for (size_t i=0; i<count; ++i)
	  ;
 //   ::operator<<(s, ref[i]);
}

template<class T>
void ACOM_NET_NAMESPACE::ACOMWrapperFormat<T>::read
  ( ACOM_NET_NAMESPACE::IStream& s, void* val, size_t count ) const
{
  T* ref = (T*)val;
  for (size_t i=0; i<count; ++i)
	  	;
//    ::operator>>(s, ref[i]);
}
#endif
