// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_iovect.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// $Id: acom_iovect.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
//
// ==========================================================================
#ifndef _ACOM_IOVECT_H_
#define _ACOM_IOVECT_H_
#include <inuk/acom_ns.h>
#ifndef WIN32
#include <inuk/acom_netconfig.h>
#endif
#include <inuk/acom_types.h>
#include <stddef.h>      // - for size_t

//@Include: acom_netconfig.h

namespace ACOM_NET_NAMESPACE {
  
// -------------------------------------------------------------------------
/** IOVectEntry
    Data structure for IOVect
*/
// -------------------------------------------------------------------------
struct IOVectEntry {
  /// Constructor of new iovect of given length
  IOVectEntry( void* d =0, size_t l =0)
  : data(d), len(l)
  {}
  void*  data;
  size_t len;
};

/** IOVectSize
*/
typedef int32_t  IOVectSize;



/*!
  \brief  Header of an IOVCell.

  An IOVCell is an iovector which may be used as an basic cell to build a
  sequence of IOvector entries. The header (static part) of the IOCell
  contains the link and the number of filled iovec entries.
  They are followed by these entries.
*/
class IOVectCell : public PTH::Linkable<IOVectCell> {
public :
  int _filled;
  IOVectEntry   _iov[1]; // first iovector entries
  IOVectCell() : _filled(0) {}
  ~IOVectCell() {}
  friend class IOVectSeq;
};

  
/*!
  \brief Allocator of fixed size IOVCell with cache for fast allocation
  (thread unsafe)

  A list registers freed IOVCell until a given theshold.
  Allocation tries to found a block in the list before using standard allocation.
  Deallocation puts a block in the list until list size is equal to threshold.
  At that time, normal deallocation is done.

  \warning : no control of size is done
*/
typedef PTH::SyncBlockAllocator<IOVectCell> IOVectCellAllocator;


/*!
  \brief Sequence of IOvector Entries

  A unlimited sequence of iovec entries is build from a single linked list of
  fixed sized IOVCell. This stream supports only one access mode at a
  time (read THEN write). Before read, IOVectCell::putEos() must be issued.
  If not, effect of following reads are undefined. It is thread unsafe.

  The total number of entries and the cumulated  number of bytes is maintained
  for a sequence.( IOVectCell::getByteSize())

  To be usable, a sequence must be bound to an IOVectCell allocator.

  \warning Today Iovector entries must write or read in sequence from the first
  to the last.
  */
class IOVectSeq {
  /* BUG in g++2.95.4: should be public for IteratorCell protected :
  */
public:
  int         _iovG;  /// total number of iovect entries in stream
  size_t      _szG ;  /// total data (byte) size in  stream( to read or writed)
  IOVectCell* _iovb;  /// current iovec cell
  int         _iovsz; /// number of entries in cell

  /** queue of iovector
  */
  PTH::List<IOVectCell> _q;

  /** IOOvec block allocator
  */
  IOVectCellAllocator* _iovA;

public:

  /** default constructor
  */
  IOVectSeq(): _iovG(0), _szG(0), _iovb(0), _iovsz(0), _q(), _iovA(0) {}

  /** initiator :
    the Cell allocator and cell size ( max number of iovevct enties) are  given
  */
  bool init(int iovsz, IOVectCellAllocator * iova )
  { _iovsz=iovsz; _iovA=iova; return true; }

  /** destructor : close and return cells to allocator
  */
  virtual ~IOVectSeq(){ }

  /** close  a  sequence after writing/reading all iovec entries ; free all cells
  */
  void destroy() {
    while(!_q.empty())  _iovA->free(_q.pull_front());
    _iovG=0;_szG=0;_iovb=0;_iovsz=0;_iovA=0;
  }


  /** Operators for building incrementally the IOVectSeq
  */
  //@{
  /** push end of sequence
    */
  void putEos(){ };


  /** pack an iovec entry(true if OK)
    */
  bool put( IOVectEntry * s)
  {
    if(!_iovb){
      _iovb= _iovA->alloc();
      if(!_iovb) return false;
      _q.push_back(_iovb); _iovb->_filled=0;
      _iovb->_iov[_iovb->_filled++]=*s;  _szG+=s->len;
    } else {
      _iovb->_iov[_iovb->_filled++]=*s;  _szG+=s->len;
      if(_iovb->_filled==_iovsz)_iovb=0; // current iovec is full
    }
    _iovG++;
    return true;
  }

  /**  pack an iovec entry(true if OK)
    */
  bool put( void * b, size_t sz)
  {
    IOVectEntry iove(b,sz);
    return put( &iove );
  }
  //@}

  /**  get cumulated number of bytes
    */
  size_t getByteSize() 
  { return _szG;}

  /**  get cumulated number of byte blocks(ie IOVect entries)
    */
  int getBlockSize() 
  { return _iovG;}


  /** Iterator through the list of IOVectCell
  */
  //@{
  class IteratorCell {
    IOVectCell * _cell;
  public:
    //typedef IOVectCell * value_type;
    IteratorCell():_cell(0){}
    IteratorCell(IOVectSeq & ios ):_cell(ios._q.front()){}
    IteratorCell(IOVectSeq * ios ):_cell(ios->_q.front()){}
    IteratorCell(IOVectCell * c ):_cell(c){}
    ~IteratorCell(){}
    IOVectCell& operator*(){return  *_cell;}
    IOVectCell* operator->(){return _cell;}
    IOVectCell* operator()(){return _cell;}
    bool operator==(const IteratorCell& it){return (_cell==it._cell);}
    bool operator!=(const IteratorCell& it){return (_cell!=it._cell);}
    //prefix operator
    IteratorCell & operator++()
    { _cell=_cell->next(); return *this; }
    friend class IteratorEntry;
  };// end class IteratorCell
  //IteratorCell beginCell()const{return IteratorCell(this);};
  //IteratorCell endCell()const{return IteratorCell();};
  IteratorCell beginCell()
  { return IteratorCell(this); }
  IteratorCell endCell()
  { return IteratorCell(); }
  //@}


  /** Iterator through the list of IOVectEntry
    */
  //@{
  class IteratorEntry {
    IteratorCell _icell;  // Cell iterator
    int _e; // entry index
  public :
    //typedef IOVectEntry * value_type;
    IteratorEntry():_icell(),_e(0){}

    IteratorEntry(IOVectSeq & ios ):_icell(ios),_e(0){}

    IteratorEntry(IOVectSeq * ios ):_icell(ios),_e(0){}

    IteratorEntry(IOVectCell * c ):_icell(c),_e(0){}

    ~IteratorEntry(){}

    IOVectEntry& operator*(){
      return _icell->_iov[_e];
    }

    IOVectEntry* operator->(){
      return &_icell->_iov[_e];
    }

    bool operator==(const IteratorEntry& it){
      return (_icell==it._icell)&&(_e==it._e );
    }

    bool operator!=(const IteratorEntry& it){
      return !((_icell==it._icell)&&(_e==it._e ));
    }

    //prefix
    IteratorEntry & operator++(){
      if((++_e)>=_icell->_filled) {_e=0; ++_icell; }
      return *this;
    }

    /** unpack iovec entry.
      return true if OK, false if  end of sequence  ( Destructive read ???)
      */
    bool get( IOVectEntry * d) {
      // std::cout << "IteratorEntry::get " << (int)(_icell()) << std::endl ;
      if( _icell() ==0 ){
        // std::cout << "????????????"<<std::endl;
        d->data=0;d->len=0;
        return false;
      }
      *d=_icell->_iov[_e];
      _e++;
      if((_e)>=_icell->_filled) { ++_icell;_e=0;}
      return true ;
    }

    /** unpack iovec entry.
      return true if OK, false if end of sequence  ( Destructive read ???)
    */
    bool get( void ** b, size_t* sz)
    {
      IOVectEntry iove;
      if (!get( &iove )) {
        *b = 0;*sz =0;
        return false;
      }
      *b = iove.data;
      *sz = iove.len;
      return true;
    }

    /** end of iovec entries stream
    */
    bool eos()
    {  return _icell() ==0 ;}
  }; // end class IteratorEntry

  IteratorEntry beginEntry()
  { return IteratorEntry(this); }
  IteratorEntry endEntry()
  { return IteratorEntry(); }
  // IteratorEntry beginEntry()const{return IteratorEntry(this);};
  // IteratorEntry endEntry()const{return IteratorEntry();};
  //@}
};



/*! \brief  Two artefact data structures that permits to have direct access
  to both  iovector with accumulation of size.

  using basis of expression template, io_v[i] = iov is evaluated as:
  -1/ io_v is an IOVector in place of a pointer ot IOVectEntry, that
  store the IOVectEntry* of the request + a pointer to the size_d
  of the network header
  -2/ io_v[i] is an IOVectEntryRef that store same thing as IOVector
  plus the index 'i'.
  -3/ io_v[i] = iov calls the method operator= in class IOVectEntryRef
  that do normal allocation of both IOVectEntry and the accumulation
  of the size.

  \warning MUST BE ADAPTED to the fact iovector need to be read by service
  handler and the use of unbounded sequence of iovec entries.
  */
class IOVector {
public:
  class IOVectEntryRef {
    IOVectEntryRef& operator=( const IOVectEntryRef& iov );
public:
    IOVectEntryRef& operator=( const IOVectEntry& iov )
    {
      _iov[_ith] = iov;
      *_cumul += iov.len;
      return *this;
    }
protected:
    friend class IOVector;
    IOVectEntryRef(int ith, IOVectEntry* iov, uint32_t* cumul)
      : _ith(ith), _iov(iov), _cumul(cumul) {}
    int          _ith;
    IOVectEntry* _iov;
    uint32_t*    _cumul;
  };

  IOVectEntryRef operator[](int i) {
    return IOVectEntryRef(i, _iov, _cumul);
  }

  IOVector( IOVectEntry* iov =0, uint32_t* c  =0)
    : _iov(iov), _cumul(c)
    {}
  IOVectEntry* _iov;
  uint32_t*    _cumul;
};


} // -namespace

#endif // _ACOM_IOVECT_H_
