// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_netclient.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Jacques Briat, Thierry Gautier, Cyrille Martin
// ==========================================================================
#ifndef _ACOM_NETCLIENT_H_
#define _ACOM_NETCLIENT_H_
#include <inuk/acom_netbase.h>
#include <inuk/acom_iovect.h>
#include <inuk/acom_stream.h>

//@Include: acom_netconfig.h acom_comerror.h acom_types.h

// ACOM namespace
namespace ACOM_NET_NAMESPACE {


/* fwd declaration */
class Callback;

// ---------------------------------------------------------------------------
// NetworkClient class
//! The virtual API for new Inuktitut interface Network (client side).
//! @ingroup NetDev
//! @ingroup NetUsers
//! The Network interface is split in serveral sub interfaces.
//! NetworkClient deals with sending message.
// ---------------------------------------------------------------------------

class NetworkClient {
public:

  //! Default Cstor
  NetworkClient();
  
  //! Destructor
  virtual ~NetworkClient();
  
  // -------------------------------------------------------------------------
  // Request class
  //! @brief Format of sent requests to the network
  //! @ingroup Request
  //! This class is the client side descriptor of a send request. 
  //! It registers the data to send and maintains the state of progress of the 
  //! sending actions.
  //! A request may be reused concurrently or sequentially.  It is garbaged 
  //! when the announced posts are all terminated. The callback is done 
  //! before the request deallocation.. 
  // ------------------------------------------------------------------------- 

  class Request : public NetworkBase::BaseRequest {  
  public :  

    //! default constructor
    Request();

     //! destructor
    ~Request();

    //! initializer of request
    //! @param bsz  ??
    //! @param bufa ??
    //! @param iovsz ??
    //! @param iova ??
    //! @param manager ??
    //! @return true if successfull
    bool init(  size_t bsz, ByteCellAllocator* bufa,
                int iovsz, IOVectCellAllocator* iova,
                Network* manager );

    //! finalizer of a request ; free all resources
    void finish();
  
    //! notify that the request has been completely sent
    //! This function should call the call back function and release the
    //! request if not needed anymore
    void notify();

    //! Pack an 'End of Stream' mark on the immediate datastream
    void packEos( );

    //! Pack some data on the immediate datastream value
    //! @param s address of data
    //! @param ssz size of data to be packed
    //! @return the size of packed data
    //! @remark Data are copied so the memory region pointed by \a s 
    //! can be freed after the call.
    size_t pack( const void* s, size_t ssz);

    //! Put an iovec entry in deferred sequence
    //! @return true if successfull
    bool put( IOVectEntry* s);

    //! Put some data an in deferred sequence
    //! @param b address of the memory region containing deferred data
    //! @param sz size of deferred data
    //! @return true if successfull
    //! @warning the memory region pointed by \a b should not be freed until 
    //! the request is completely sent, that is when the done function of 
    //! the callback object is called by the network
    bool put( void * b, size_t sz);

    //! Get the immediate datastream
    ByteSeq* get_datastream();

    //! Get the deferred IOvect sequence
    IOVectSeq* get_iovstream();

    //! Return a pointer to the extended Network header
    void* get_ext_header();
    
  public:
    //! request state (??)
    //! default 1 when request is created.
    //! -1 when request is completed. 
    //! Then 0 callback is called
    PTH::AtomicCounter _used_ct ; 
    
    //! Callback to call at local completion date.
    //! The done function is to be called on caller 
    //! by the network when the request has been completely sent to all nodes
    //! @warning Until this function is called, none of the deferred data
    //! should be dealocated. Once it has been called, none of the deferred 
    //! data should be accessed
    Callback* _f_cbk;   
    
    ByteSeq   _datastream;  //!< stream of immediate data
    IOVectSeq _iovstream;   //!< stream of deferred data blocks
  };



  //! @defgroup Channel 
  //! @brief Channel related types and functions
  //! @ingroup NetDev
  //! @ingroup NetUsers
  //! @{
  // -------------------------------------------------------------------------
  //! OutChannel class
  //! @brief Communication channel abstraction used to send requests
  //! 
  //! The implementation of a specific OutChannel for a particular network
  //! should be done by making a derivate class of OutChannel for that network.
  //! The post method should be redefined
  //! @see Network
  //! @see ProtocolService
  // -------------------------------------------------------------------------
  class OutChannel {
  public:
    //! Default constructor
    OutChannel();
    
    //! Destructor 
    virtual ~OutChannel();
        
    //! Post a request on the channel
    //! @param req request to post
    virtual void post( Request* req ) throw (ComFailure)=0;
        
    //! Return the network to which belongs the channel
    Network* get_net();

    //!  Return the identity of the peer node
    NetworkBase::LocalNodeNo get_peer();
    
  protected: 
    //! Initilize a channel 
    //! @param net network on which the channel must be initilized
    //! @param peer node peer to which the channel must be established
    void init(Network* net, NetworkBase::LocalNodeNo peer);

  protected: 
    Network*                 _net;  //!< medium for communication 
    NetworkBase::LocalNodeNo _peer; //!< node with remote connection
    friend class Network;
  }; // -- end class Channel
  
  // @} // group channel


  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  //! @name Request management
  //@{ 
  //! Allocate a request to be sent on this network
  //! This function is called at request creation time (??)
  //! @return a pointer to the allocated request
  virtual Request* allocate_request ( ) throw (BadAlloc) =0;

  //! deallocate a request
  //! @param r the request to be deallocated
  virtual void deallocate_request( Request* r ) throw (InvalidArgument) =0;

  //! Cancel a request
  //! @param req the quest to be cacelled
  //! @remark Cancelation of request depend on state of progress
  //! @return true if successfull
  virtual bool cancel_request( Request* req ) 
    throw (InvalidArgument);

  //! send  to a node (AM,AWS,WS protocols) 
  //! @param   req  request (IN) 
  //! @param   ch  channel  (IN) 
  virtual void post( Request* req, OutChannel* ch=0 ) 
       throw (ComFailure);

  //@} 

protected:
  friend class BaseRequest;
  friend class OutChannel;
  friend class Request;
  friend class Receipt;
}; // -- end class Network


/*
 * inline definition
 *
 */
// ---------------------------------------------------------------------------
// Request class
// --------------------------------------------------------------------------- 
inline NetworkClient::Request::Request()
  : NetworkBase::BaseRequest(), _used_ct(0),
    _f_cbk(0), _datastream(), _iovstream() 
{}

inline NetworkClient::Request::~Request() 
{}

inline bool NetworkClient::Request::init(
  size_t bsz, ByteCellAllocator* bufa,
  int iovsz, IOVectCellAllocator* iova,
  Network* manager 
)
{ 
  NetworkBase::BaseRequest::init(manager);
  if (!_iovstream.init( iovsz, iova )) 
    return false;
  if (!_datastream.init( bsz, bufa, iovsz, iova))
    { _iovstream.destroy(); return false; }
  return true;
}

inline void NetworkClient::Request::finish() 
{ 
  _state = NetworkBase::S_DESTROY; _error_no = ComFailure::OK;
  _datastream.destroy(); _iovstream.destroy(); 
}

inline void NetworkClient::Request::packEos( )
{ _datastream.packEos(); }

inline size_t NetworkClient::Request::pack( const void* s, size_t ssz)
{ return _datastream.pack( s, ssz ); }

inline bool NetworkClient::Request::put( IOVectEntry* s) 
{ return _iovstream.put(s); }

inline bool NetworkClient::Request::put( void * b, size_t sz) 
{ return _iovstream.put(b,sz); }

inline ByteSeq* NetworkClient::Request::get_datastream() 
{ return &_datastream; }

inline IOVectSeq* NetworkClient::Request::get_iovstream() 
{ return &_iovstream; }

inline void* NetworkClient::Request::get_ext_header()
{ return _ext_header; }


// ---------------------------------------------------------------------------
// Channel class
// ---------------------------------------------------------------------------
inline NetworkClient::OutChannel::OutChannel() : _net(0) 
{}

inline NetworkClient::OutChannel::~OutChannel() 
{}
    
inline Network* NetworkClient::OutChannel::get_net() 
{ return _net; }

inline NetworkBase::LocalNodeNo NetworkClient::OutChannel::get_peer() 
{ return _peer; }

inline void NetworkClient::OutChannel::init(Network* net, NetworkBase::LocalNodeNo peer)
{ _net=net; _peer=peer; }



} // end namespace ATH...
#endif 
