// ==========================================================================
// $Source: /home/cvs/Taktuk2/NetworkCore/src/acom_system.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// $Id: acom_system.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
#ifndef _ACOM_SYSTEM_H_ 
#define _ACOM_SYSTEM_H_

#include <inuk/acom_netconfig.h>
#include <inuk/acom_error.h>
#include <inuk/acom_network.h>
#include <inuk/acom_format.h>
#include <inuk/acom_argdispatcher.h>
extern "C" {
#include <stddef.h> // for size_t
#include <unistd.h> // pid_t
}
#include <map>

//@Include: acom_config.h acom_error.h

namespace ACOM_NET_NAMESPACE {

// ------
// System errors: definition exceptions throwed by this package.
// ------
//@{
///Abort exception throwed to abort the program 
DERIVED_EXCEPTION ( ErrorAbort,RuntimeError); 
//@}



// ------
// System class: definition of initialisation/finalization 
// of Athapascan virtual machine.
// ------
/** System.
Definition of functions (static methods) to initialize and to finalize
or abort the virtual machine of Athapascan0. 
\verbatim
#include "a0++.h"

using namespace a0;

int main(int argc, char** argv)
{
  System::init(&argc, &argv);
  ...
  std::cout << "The number of nodes is = " << System::node_count << endl;
  System::initcommit();
  doCompute();
  System::end();
}
\endverbatim
*/
class Acom {
public:
  /** Abortion.
    A call to this method aborts the execution of the parallel program
    and return the error code err. The method throw an exception
    that cannot be catched (exception by a clause like 'catch(...) { }').
    @param err the error code
    @exception ErrorAbort
  */
  static void abort( int err ) throw (ErrorAbort);

  /** Initialization.
    A call to this method should be executed in order to pass the 
    first stage of the initialization process of the virtual machine.
    This method must be called before the use of any
    Athapascan functions, especially if they occurs in initialization
    of static objects. See Module class and methods 
    to provide a secure initialization/finalization scheme.
    @param argc a pointer to the number of arguments 
    @param argv a pointer to the arguments 
    @exception ErrorAbort
    @exception InvalidArgument
    @see Module
  */
  static void initialize(int* argc, char*** argv) 
     throw (ErrorAbort,InvalidArgument);

  /** Initialization.
    A call to this method should be executed in order to pass the second 
    and last stage of the initialization process of the virtual machine.
    This function must be called after global declarations or allocations 
    of objects/types/ressources. 
    In return to this function call, each node has reached the same point and
    all the modules on each node are initialized.  
    See Module class and methods to provide a secure initialization/
    finalization scheme.
    @exception ErrInitCommit
    @see Module
  */
  static void commit() throw ();

  /** Finalization.
    This function must be called at the end of the application.
    After the call of this function, each node has all of its modules
    finalized.
  */
  static void terminate() throw (ErrorAbort);

  /** Return the number of network on which the current node is connected to.
  */
  static int network_count();
  
  /** Bind a name to a new network
  */
  static void bind_network( const std::string& name, Network* net )
    throw (InvalidArgument);

  /** ReBind a name to a network
  */
  static void rebind_network( const std::string& name, Network* net )
    throw (InvalidArgument);

  /** Resolv the name for a network
  */
  static Network* resolv_network( const std::string& name )
    throw (InvalidArgument);
  
  /** Unbind a name 
  */
  static void unbind_network( const std::string& name )
    throw (InvalidArgument);

  /** Return the default network
  */
  static Network* get_default_network( );  

  /** Return an (forward) iterator over all connected networks to this node
  */
  //@{
  typedef std::map<std::string, Network*> AllNetworks;
    ///
  struct const_iterator: public AllNetworks::const_iterator {
    const_iterator( AllNetworks::const_iterator it )
     : AllNetworks::const_iterator(it)
    {}
    Network* operator*() const { 
      return  AllNetworks::const_iterator::operator*().second;
    }
  };
    ///
  static const_iterator	begin_network();
    ///
  static const_iterator	end_network();  
  //@}

  /**@name release and license informations */
  //@{
  /// Return the license information to use this library.
  static const char* info();

  /// Return the version and release number.
  static const char* version();

  /// Return the recette of the 'gateau au chocolat'.
  static const char* gateau_chocolat();
  //@}

  /**@name attach a debugger on the process that call it */
  static void attach_debugger() throw (Runtime);

  /**@name return the canonical name of the machine
  */
  static const std::string& get_canonical_name();

  /**@name return an unique identifier/canonical name of the local processus. 
  */
  static pid_t get_pid();

  /**@name return the size of a system page_size */
  static size_t get_pagesize();

  /**@name Get access to argv, argc.
     getargv return the i-th argument of the programm.
     getargc return the total current number of arguments.
  */
  //@{
    ///
  static char* get_argv(int i) throw (InvalidArgument);
    ///
  static int get_argc();
  //@}


protected: 
  // - argv/argc of the program
  static int*    argc;
  static char*** argv;

public:
  /** Export arg and parsing method 
  */
  static ACOM::Arg* arg;
  static ACOM::Arg::ArgInfo* unparsed_args_info;
  static ACOM::Arg::ArgInfo* all_args_info;

private:
  friend class SystemInit;

public: // - public because call by external C function
  static void sv_signal_hanler(int v); // - debug of A0 

protected:
  friend class NetworkBase;
  
  /// all registered newtork
  static AllNetworks all_networks;
  static std::vector<Network*> all_nids;
};


} // -namespace

#endif
