// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_error.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat T. Gautier
// [note]: class hierarchy for Inuktitut0 errors is derived from ANSI/ISO C++ hierarchy.
// $Id: pth_error.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $
// ==========================================================================
/*!
\file  
\brief Exception  hierarchy for Inuktitut0 INRIA.
\author  J. Briat T. Gautier 
\date 2001

The  Inuktitut0 exception  hierarchy  try to match the ANSI/ISO C++ hierarchy.
It is extended with Posix exceptions associated with Pthreads 
*/



//
// ==========================================================================
#ifndef _PTH_ERROR_H_ 
#define _PTH_ERROR_H_

// the namespace identifier
#include <inuk/pth_ns.h>
#include <inuk/pth_dep.h>
// the root of standard std::exception hierarchy
#include <exception>   
// - the standard  hierarchy 
#include <stdexcept>   
// - for bad_alloc
#include <new>       
extern "C" {  
// - for system error
#include <errno.h>
// for system error messages
#include <stdio.h>
// for strerror
#include <string.h> 
}
#include <string>   

 
namespace PTH
{

//-------------------------------------------------------------------------
// errors for Pthread package are instance of RuntimeError
//-------------------------------------------------------------------------
/** 
\brief  RuntimeError is the base class to catch/throw exceptions for runtime unix/posix errors

Derived from  the ANSI/C++  exception hierarchy.
\code
  class exception
  {
  public:
    exception() throw() { }
    virtual ~exception() throw();
    virtual const char* what() const throw();
  };  
\endcode

 We can use the macro  DERIVED_EXCEPTION() to generate a new derived exception and the macros  
ASSERT() or PTH_ASSERT() to throw an exception if a condition is false.

*/
class RuntimeError: public  std::exception {
  ///  function to translate  a system error code to a string
  virtual const char * _whatmsg( ) const
  { return strerror(_code);}

public :
  /// default constructor 
  RuntimeError() throw() : std::exception(), _code(0), _msg(0){}
  /// constructor : integer code + message
  RuntimeError(const char* msg , int code=0) throw()
  : std::exception(),_code(code),_msg(msg){}
  /// constructor : integer code 
  RuntimeError(int code ) throw() 
  : std::exception(), _code(code), _msg( "[Runtime OS Error]")
  {}

  /// function to write the exception into the stream o
  std::ostream& print( std::ostream& o ) const;
  
  /// function to build a full error message
  const char * what() const throw ();

  /// function return the error code
  int code() const throw ();
     
protected :
    int _code;
    const char * _msg;
};

 /**@name Static methods to throw an exception.
   The purpose of this method is to be able to set a breakpoint on throw 
   @param err a reference to an Exception object
   @exception the err object
 
  void throwit( const RuntimeError& err ) throw(const RuntimeError&);
 */
extern void throwit(  const RuntimeError& err ) throw(const RuntimeError&);


/** 
\brief  Macro to generate error class inherited from RuntimeError
    
*/
#define DERIVED_EXCEPTION( E1,E0)\
    class E1: public  E0{\
       public :\
        E1( const char * msg , int code=0) : E0(msg,code){};\
        E1 (int code=0 ) : E0(""#E1"",code){};\
  };



/** 
\brief   exception for posix thread errors
*/
class PosixError : public RuntimeError{
public :
  PosixError (const char * msg , int code=0): RuntimeError(msg,code){}
  PosixError ( int code=0): RuntimeError("PosixError",code){}
};


/** 
\brief   exception for IO errors
*/
class IOError : public RuntimeError{
public :
  IOError (const char * msg , int code=0): RuntimeError(msg,code){}
  IOError ( int code=0): RuntimeError("IOError",code){}
};


/** 
\brief  NoMoreResource  exception generated when detecting a 
lack of ressources 

*/
class NoMoreResource : public RuntimeError{
public :
  NoMoreResource(const char * msg , int code=0): RuntimeError(msg,code){}
  NoMoreResource( int code=0): RuntimeError("NoMoreResource",code){}
};
  

/** BadAlloc. 
*/
class BadAlloc : public RuntimeError{
public :
  BadAlloc(const char * msg , int code=0): RuntimeError(msg,code){}
  BadAlloc( int code=0): RuntimeError("BadAlloc",code){}
};


/** Range. An object of this class is throwed if
    an un-previsible runtime error occurs while checking a 
    domain for a value.
*/
DERIVED_EXCEPTION(Range, RuntimeError );

/** Logic. An object of this class is throwed for
    error that could be, in principle, detected at compilation.
*/
DERIVED_EXCEPTION(Logic, RuntimeError );

/** InvalidArgument. An object of this class is throwed if
    the argument of a call to a function is invalid.
*/
DERIVED_EXCEPTION(InvalidArgument,Logic);


/** Length. 
*/
DERIVED_EXCEPTION(Length, Logic);

/** OutOfRange. 
*/
DERIVED_EXCEPTION(OutOfRange,Logic);


/*! \def ASSERT(msg)
\brief  throws the \b EXCPT exception if \b cond is false 
*/  
#define ASSERT( cond,EXCPT)   if (!(cond)) throwit (EXCPT);

/*! \def PTH_ASSERT(msg)
\brief  throws the \b EXCPT exception if \b cond is false 
*/  
#define PTH_ASSERT( cond,EXCPT)   if (!(cond)) throwit (EXCPT);

/*! \def ASSERT_DEBUG(msg)
\brief  throws the \b EXCPT exception if \b cond is false 
*/
#ifdef DEBUG
#  define ASSERT_DEBUG( cond,EXCPT)
#else
#  define ASSERT_DEBUG( cond,EXCPT)   if (!(cond)) throwit (EXCPT);
#endif

/*! \def PTH_ASSERT(msg)
\brief  throws the \b EXCPT exception if \b cond is false 
*/
#ifdef DEBUG
#  define PTH_ASSERT_DEBUG( cond,EXCPT)
#else
#  define PTH_ASSERT_DEBUG( cond,EXCPT)   if (!(cond)) throwit (EXCPT);
#endif


} // - namespace PTH_...

#endif
