// ==========================================================================
// $Source: /home/cvs/Taktuk2/Pthread/src/pth_util.h,v $ 
//Copyright(c)'2001 by Apache Team // see the copyright file 
// $Id: pth_util.h,v 1.1.1.1 2004/06/29 13:32:12 martinc Exp $ 
// 
// 
//
//==========================================================================

/*!
\file  
\brief  Posix thread for Inuktitut0++ INRIA 
\author  J. Briat T. Gautier 
\date 2001

This file contains the wrappers to :
  - posix locks PTH::Semaphore
  - and other utility functions

It provides operators on these objects throwing Posix Exceptions.
\sa pth_error.h PTH::PosixError
*/
#ifndef _PTH_UTIL_H_
#define _PTH_UTIL_H_

// the namespace identifier
#include <inuk/pth_ns.h>
#include <inuk/pth_dep.h>
#include <inuk/pth_macro.h>
#include <inuk/pth_atomic.h>
#include <inuk/pth_pthreadsync.h>


namespace PTH { 

// -----
// CountDown
// -----
/**
\brief This class implements interface to Count down counter
*/

class CountDown {
public:
    /// Initialize event. Initial value is 0.
    CountDown() throw (PosixError);    
    void init(int i) throw ( PosixError );
    /// Destuctor. 
    virtual ~CountDown() throw( PosixError){};
    void destroy() throw ( PosixError );
    /// decrement by one the counter. Result is the new value
    int operator--();
    /// increment by one the counter. Result is the new value
    int operator++(){CRITICAL(_mutex); ++_ct; return _ct;};
    /// get counter. 
    int getValue(){return _ct;};
    /// wait for counter becoming <=0
    void wait() throw (PosixError);
private:
    // do not need an Atomic counter any more with the new synchronization scheme
    volatile int _ct; 
    //    AtomicCounter _ct;
    /// lock needed for exclusive access
    Mutex  _mutex;
    /// condition to wait for positive value
    Condition  _cond;
};
// -----
// CountDown
// -----
/**
\brief This class implements interface to Count down counter
*/

class NonStopCountDown {
public:
    /// Initialize event. Initial value is 0.
    NonStopCountDown() throw (PosixError);    
    void init(int i) throw ( PosixError );
    /// Destuctor. 
    virtual ~NonStopCountDown() throw( PosixError){};
    void destroy() throw ( PosixError );
    /// decrement by one the counter. Result is the new value
    int operator-=(int i);
    /// increment by one the counter. Result is the new value
    int operator+=(int i);
    /// get counter. 
    int getValue(){return _ct;};
    /// wait for counter becoming <=0
    void wait() throw (PosixError);
private:
    volatile int _ct; 
    /// lock needed for exclusive access
    Mutex  _mutex;
    /// condition to wait for positive value
    Condition  _cond;
};

// -----
// PersistentEvent
// -----
/**
\brief This class implements interface to Persistent event.

When set, event remains set until explicitly reset
*/

class PersistentEvent
{
public:
    /// Initialize event. Initial value is reset.
    PersistentEvent() throw (PosixError); 
	 /// Initialize event. Initial value is reset.  
    void init() throw ( PosixError );

    /// Destuctor. 
    virtual ~PersistentEvent() throw( PosixError){};

    /// set event; wake up all waiting threads; event remains set 
    void set() throw (PosixError);
    /// set event; wake up all waiting threads; event is reset 
    void pulse() throw (PosixError);
    /// reset event
    void reset() throw (PosixError);
    /// wait event set
    void wait() throw (PosixError);
    /// timed wait event set
    void wait(long milli, int nano=0) throw (PosixError);

private:
    volatile bool _set;
    /// lock needed for exclusive access
    Mutex  _mutex;
    /// condition to wait for positive value
    Condition  _cond;
};

// -----
// TransientEvent
// -----
/**
\brief This class implements interface to transient event.

When set, event remains set until one thread is wake up
*/
class TransientEvent {
public:
  /// Initialize event. Initial value is reset.
  TransientEvent() throw (PosixError);  
  /// Initialize event. Initial value is reset.
  void init() throw ( PosixError );

  /// Destuctor.
  virtual ~TransientEvent() throw( PosixError){};

  /// set event; reset only when a thread may be awaken
  void set() throw (PosixError);
  /// set event; wake one waiting thread if any; event is allways reset 
  void pulse() throw (PosixError);
  /// reset event
  void reset() throw (PosixError);
  /// wait event set
  /// wait event set
  void wait() throw (PosixError);
  /// timed wait event set
  void wait(long milli, int nano=0) throw (PosixError);
private : 
  volatile bool _set;
  volatile int  _ct; /// waiting threads 
  /// lock needed for exclusive access
  Mutex  _mutex;
  /// condition to wait for positive value
  Condition  _cond;
};


// -----
// Semaphore
// -----
/**
\brief This class implements interface to Dijkstra semaphore
   based on Posix thread.

This kind of semaphore must be initialized with positive values.
*/

class Semaphore
{
public:
  /// Initialize the Semaphore. Initial value must not be negative.
  Semaphore() throw (PosixError) 
   :  _mutex(), _cond(), _counter(0)
  {}

  void init(int i) throw ( PosixError )
  { _counter=i;_mutex.init();_cond.init(); 
    ASSERT( (_counter>= 0), 
      PosixError( SUFX("[Semaphore::init] initial value must be >= 0"),0) );
  }

  /// Destuctor. Throw an error if there is waiting threads
  virtual ~Semaphore() throw( PosixError);

  /// Decrement the semaphore and block if it is <= zero
  void p() throw (PosixError);

  /// Increment the semaphore. Unblock one thread if it becomes >0
  void v() throw (PosixError);

private:
  /// lock needed for exclusive access
  Mutex        _mutex;
  /// condition to wait for positive value
  Condition    _cond;
  /// count of V done  minus count of P done
  volatile int _counter;
};



//******************************************************************
//******************************************************************



// \ingroup Glib
/*!    \brief Allocator of fixed size objects  with cache for fast allocation   (thread safe)
A list  registers freed blocks until a given theshold. 
Allocation tries to found a block in the list before using standard allocation.
Deallocation puts a block in the list until list size is equal to threshold. At that time
normal deallocation is done.
We can define a  new allocator  
\code 
SyncCellAllocator<X> A_of_X;
\endcode  
*/

template < class T > 
class SyncCellAllocator: public CellAllocator<T>    {
public:
  /// lock needed for exclusive access
  Mutex          _mutex;
public:
  /** Constructor with a given threshold
  */
  SyncCellAllocator (int v):CellAllocator<T>(v),_mutex()
  { _mutex.init(); }

  /** default Constructor with very big threshold
  */
  SyncCellAllocator ():CellAllocator<T>(),_mutex()
  { _mutex.init(); }

  /** \briefdestructor
      free all cached blocks 
  */
  ~SyncCellAllocator() { _mutex.destroy(); }
  
  /** init with a given threshold 
  */
  void init(int v){ CellAllocator<T>::init(v);}

  /** \brief Alloc a block of type T
      tries to get it from cached list else normal allocation
  */
  T* alloc(){
    CriticalGuard gd(_mutex);
    return CellAllocator<T>::alloc() ;
  }

  /** \brief Free a block of type T
      put in cached blocks if cache size is less than threshold else deallocate 
  */
  void free( T* b){
    CriticalGuard gd(_mutex);
    CellAllocator<T>::free(b);
  }
};



//******************************************************************
//******************************************************************
/*! 
\brief Allocator of fixed size block with cache for fast allocation (thread safe)
An allocator manages blocks of constantt size. This size is fixed when an allocator 
object is build. A list registers freed blocks until a given theshold.  
Allocation tries to found a block in the list before using standard allocation. 
Deallocation puts a block in the list until list size is equal to threshold. 
At that time, normal deallocation is done.

Warning : no control of size is done
*/

template < class T >
class SyncBlockAllocator : public BlockAllocator<T>  {
public:
  Mutex          _mutex;
public:
  SyncBlockAllocator (): BlockAllocator<T>(), _mutex() { _mutex.init(); }
  /// Constructor with a given threshold
  SyncBlockAllocator (size_t bsz, int v) : BlockAllocator<T>(bsz,v), _mutex() { _mutex.init(); }
  /// default Constructor with very big threshold
  SyncBlockAllocator (size_t bsz): BlockAllocator<T>(bsz), _mutex() { _mutex.init(); }

  /** \brief  destructor
      free all cached blocks 
  */
  ~SyncBlockAllocator (){  _mutex.destroy(); }
       
  /// init with a given threshold and size
  void init(size_t bsz, int v){ BlockAllocator<T>::init(bsz,v); }

  /** \brief Alloc a block of type IOVectBuffer
      tries to get it from cached list else normal allocation
  */
  T * alloc(){/* Get the mutex and release it at the end */
    CriticalGuard gd(_mutex);
    return BlockAllocator<T>::alloc() ;
  };

  /** \brief Free a block 
      put in cached blocks if cache size is less than threshold else deallocate 
  */
  void free( T *b){ 
    CriticalGuard gd(_mutex);
    BlockAllocator<T>::free(b) ;
  }
};


/*! 
   \brief Allocator of fixed size byte block  with cache for fast allocation   (thread safe)

An allocator manages blocks of constantt size. This size is fixed when an allocator object is build. A list registers freed blocks until a given theshold.  Allocation tries to found a block in the list before using standard allocation. Deallocation puts a block in the list until list size is equal to threshold. At that time, normal deallocation is done.

Warning : no control of size is done

*/


class SyncByteAllocator : public ByteAllocator  {
 Mutex          _mutex;
public :
   SyncByteAllocator ():ByteAllocator (),_mutex(){_mutex.init();};
   /** Constructor with a given threshold
    */
   SyncByteAllocator (size_t bsz,int v): ByteAllocator (bsz,v),_mutex()
   {_mutex.init();}

   /** default Constructor with very big threshold
    */
   SyncByteAllocator (size_t bsz): ByteAllocator (bsz),_mutex(){_mutex.init();}

   /** \brief  destructor
    free all cached blocks 
   */
   ~SyncByteAllocator  (){ _mutex.destroy();}

  /** init with a given threshold and size
   */
  void init(size_t bsz,int v) { _bsz=bsz;_limit=v;_ct=0;}

   /** \brief Alloc a block of type IOVectBuffer
      tries to get it from cached list else normal allocation
   */
   void * alloc(){ 
      CriticalGuard gd(_mutex);return ByteAllocator::alloc();
  }
  /**  \brief Free a block 
    put in cached blocks if cache size is less than threshold else deallocate 
   */
   void free( void *b){    
     CriticalGuard gd(_mutex);
     ByteAllocator::free(b) ;
   };
};





// -----
// CountDown
// -----
inline  CountDown::CountDown() throw (PosixError)
  :_ct(-1),_mutex(),_cond()
{};

inline    void CountDown::init(int i=0) throw ( PosixError )
{
  _mutex.init();_cond.init(); _ct = i ;
}
 
inline void CountDown::destroy() throw ( PosixError )
{
  _mutex.destroy(); _cond.destroy();
}

inline int CountDown::operator--()
{ 
  int x;
  CriticalGuard gd(_mutex);
  if (_ct ==0) return 0;
  if ((x=--_ct)==0){
    _cond.broadcast();
  }
  return x;
} 

inline void CountDown::wait() throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if(_ct ==0) return ;
  { 
    _cond.wait(_mutex);
  }
}
 
// -----
// NonStopCountDown
// -----
inline  NonStopCountDown::NonStopCountDown() throw (PosixError)
  :_ct(-1),_mutex(),_cond()
{}

inline    void NonStopCountDown::init(int i=0) throw ( PosixError )
{
  _mutex.init();_cond.init(); _ct = i ;
}
 
inline void NonStopCountDown::destroy() throw ( PosixError )
{
  _mutex.destroy(); _cond.destroy();
}

inline   int NonStopCountDown::operator-=(int i)
{
  CriticalGuard gd(_mutex);
  _ct -= i;
  if ((_ct)<=0){
    _cond.broadcast();
  }
  return _ct;
}

inline   int NonStopCountDown::operator+=(int i)
{
  CriticalGuard gd(_mutex);
  _ct += i;
  return _ct;
}

inline void NonStopCountDown::wait() throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if(_ct <= 0 )return ;
  _cond.wait(_mutex);
}
 

// -------------------------------------------------------------------------
// Persistent Event 
// -------------------------------------------------------------------------
inline PersistentEvent::PersistentEvent() throw (PosixError) 
  :_set(false), _mutex(), _cond()
{}

inline void PersistentEvent::init() throw ( PosixError ) 
{ _mutex.init();
  _cond.init(); 
}

inline void PersistentEvent::set()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  _set=true; _cond.broadcast();
}

inline void PersistentEvent::pulse()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  _cond.broadcast();_set=false;
}
   

inline void PersistentEvent::reset()throw (PosixError)
{ _set=false; }


inline void PersistentEvent::wait()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if (!_set)_cond.wait(_mutex);
}

inline void PersistentEvent::wait(long milli, int nano)throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if (!_set) _cond.wait(_mutex,milli,nano);
}


// -------------------------------------------------------------------------
// Transient Event 
// -------------------------------------------------------------------------
inline TransientEvent::TransientEvent() throw (PosixError) 
 : _set(false), _ct(0),_mutex(), _cond()
{}

inline void TransientEvent::init() throw ( PosixError ) 
{
  _mutex.init();
  _cond.init(); 
}

inline void TransientEvent::set()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if(_ct==0)_set=true;
  else {_cond.signal();}
}

inline void TransientEvent::pulse()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if(_ct!=0) _cond.signal();
}

inline void TransientEvent::reset()throw (PosixError)
{ _set=false; }

inline  void TransientEvent::wait()throw (PosixError)
{
  CriticalGuard gd(_mutex);
  if(!_set){  _ct++; _cond.wait(_mutex); _set=false;_ct--; }
}

inline void TransientEvent::wait(long milli, int nano)throw (PosixError)
{
  CriticalGuard gd(_mutex); 
  if(!_set){  _ct++; _cond.wait(_mutex,milli,nano); _set=false;_ct--;}
}


// -------------------------------------------------------------------------
// Sem
// -------------------------------------------------------------------------
inline Semaphore::~Semaphore()throw (PosixError)
{
  /* Get the mutex and release it at the end */
  { CriticalGuard gd(_mutex);
    ASSERT( (_counter>= 0), PosixError(SUFX("[Semaphore::destroy] there are waiting threads"),0));
  }
  _mutex.destroy();
  _cond.destroy();    
}

inline void Semaphore::p()throw( PosixError)
{
  /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);

  /* Wait on the condition if counter  less or equal 0*/
  _counter--;
  if (_counter < 0 )  _cond.wait(_mutex);
}

inline void Semaphore::v()throw (PosixError)
{
  /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
  /* Increment the counter */
  _counter ++;
  /* Signal the condition variable (no broadcast needed here because at most one blocked thread can be released). */
  if( _counter<=0)_cond.signal();
}


} // end namespace
#endif // _PTH_UTIL_H_
